unet: fix bug and coding style (#1171)

* unet: fix bug

* update readme

* remove useless code
This commit is contained in:
Wang Xinyu 2022-12-13 22:37:41 +08:00 committed by GitHub
parent f98182e1b1
commit cb9efbdf21
5 changed files with 282 additions and 424 deletions

View File

@ -13,8 +13,8 @@ include_directories(/usr/local/cuda/include/)
link_directories(/usr/local/cuda/lib64/)
# tensorrt
include_directories(/workspace/TensorRT-8.4.1.5/include/)
link_directories(/workspace/TensorRT-8.4.1.5/lib/)
include_directories(/workspace/TensorRT-7.2.3.4/include/)
link_directories(/workspace/TensorRT-7.2.3.4/lib/)
# opencv library
find_package(OpenCV)

View File

@ -1,68 +1,61 @@
# UNet
This is a TensorRT version UNet, inspired by [tensorrtx](https://github.com/wang-xinyu/tensorrtx) and [pytorch-unet](https://github.com/milesial/Pytorch-UNet).<br>
You can generate TensorRT engine file using this script and customize some params and network structure based on network you trained (FP32/16 precision, input size, different conv, activation function...)<br>
# Requirements
Pytorch model from [Pytorch-UNet](https://github.com/milesial/Pytorch-UNet).
TensorRT 7.x or 8.x (you need to install tensorrt first)<br>
Python<br>
opencv<br>
cmake<br>
## Contributors
# Train .pth file and convert .wts
<a href="https://github.com/YuzhouPeng"><img src="https://avatars.githubusercontent.com/u/13601004?v=4?s=48" width="40px;" alt=""/></a>
<a href="https://github.com/East-Face"><img src="https://avatars.githubusercontent.com/u/35283869?v=4s=48" width="40px;" alt=""/></a>
<a href="https://github.com/irvingzhang0512"><img src="https://avatars.githubusercontent.com/u/22089207?s=48&v=4" width="40px;" alt=""/></a>
<a href="https://github.com/wang-xinyu"><img src="https://avatars.githubusercontent.com/u/15235574?s=48&v=4" width="40px;" alt=""/></a>
## Create env
## Requirements
Please use TensorRT 7.x.
There is a bug with TensorRT 8.x, we are working on it.
## Build and Run
1. Generate .wts
```
pip install -r requirements.txt
```
## Train .pth file
Train your dataset by following [Pytorch-UNet](https://github.com/milesial/Pytorch-UNet) and generate .pth file.<br>
Please set bilinear=False, i.e. `UNet(n_channels=3, n_classes=1, bilinear=False)`, because TensorRT doesn't support Upsample layer.
## Convert .pth to .wts
```
cp tensorrtx/unet/gen_wts.py Pytorch-UNet/
cp {path-of-tensorrtx}/unet/gen_wts.py Pytorch-UNet/
cd Pytorch-UNet/
python gen_wts.py
wget https://github.com/milesial/Pytorch-UNet/releases/download/v3.0/unet_carvana_scale0.5_epoch2.pth
python gen_wts.py unet_carvana_scale0.5_epoch2.pth
```
# Generate engine file and infer
Build:
2. Generate TensorRT engine
```
cd tensorrtx/unet/
mkdir build
cd build
cmake ..
make
cp {path-of-Pytorch-UNet}/unet.wts .
./unet -s
```
Generate TensorRT engine file:
```
unet -s
```
Inference on images in a folder:
```
unet -d ../samples
```
# Benchmark
the speed of tensorRT engine is much faster
pytorch | TensorRT FP32 | TensorRT FP16
---- | ----- | ------
816x672 | 816x672 | 816x672
58ms | 43ms (batchsize 8) | 14ms (batchsize 8)
# test img
3. Run inference
```
wget https://raw.githubusercontent.com/wang-xinyu/tensorrtx/f60dcc7bec28846cd973fc95ac829c4e57a11395/unet/samples/0cdf5b5d0ce1_01.jpg
./unet -d 0cdf5b5d0ce1_01.jpg
```
# Further development
1. add INT8 calibrator<br>
2. add custom plugin<br>
4. Check result.jpg
<p align="center">
<img src="https://user-images.githubusercontent.com/15235574/207358769-dacf908e-f65d-4b2e-bc53-4fa2a9114c2a.jpg" height="360px;">
</p>
# Benchmark
Pytorch | TensorRT FP32 | TensorRT FP16
---- | ----- | ------
816x672 | 816x672 | 816x672
58ms | 43ms (batchsize 8) | 14ms (batchsize 8)
## More Information
See the readme in [home page.](https://github.com/wang-xinyu/tensorrtx)

View File

@ -6,7 +6,6 @@
#include <sstream>
#include <vector>
#include <opencv2/opencv.hpp>
#include <dirent.h>
#include "NvInfer.h"
#define CHECK(status) \
@ -95,46 +94,5 @@ IScaleLayer* addBatchNorm2d(INetworkDefinition *network, std::map<std::string, W
return scale_1;
}
ILayer* convBlock(INetworkDefinition *network, std::map<std::string, Weights>& weightMap, ITensor& input, int outch, int ksize, int s, int g, std::string lname) {
Weights emptywts{DataType::kFLOAT, nullptr, 0};
int p = ksize / 2;
IConvolutionLayer* conv1 = network->addConvolutionNd(input, outch, DimsHW{ksize, ksize}, weightMap[lname + ".conv.weight"], emptywts);
assert(conv1);
conv1->setStrideNd(DimsHW{s, s});
conv1->setPaddingNd(DimsHW{p, p});
conv1->setNbGroups(g);
IScaleLayer* bn1 = addBatchNorm2d(network, weightMap, *conv1->getOutput(0), lname + ".bn", 1e-3);
// hard_swish = x * hard_sigmoid
auto hsig = network->addActivation(*bn1->getOutput(0), ActivationType::kHARD_SIGMOID);
assert(hsig);
hsig->setAlpha(1.0 / 6.0);
hsig->setBeta(0.5);
auto ew = network->addElementWise(*bn1->getOutput(0), *hsig->getOutput(0), ElementWiseOperation::kPROD);
assert(ew);
return ew;
}
int read_files_in_dir(const char *p_dir_name, std::vector<std::string> &file_names) {
DIR *p_dir = opendir(p_dir_name);
if (p_dir == nullptr) {
return -1;
}
struct dirent* p_file = nullptr;
while ((p_file = readdir(p_dir)) != nullptr) {
if (strcmp(p_file->d_name, ".") != 0 &&
strcmp(p_file->d_name, "..") != 0) {
//std::string cur_file_name(p_dir_name);
//cur_file_name += "/";
//cur_file_name += p_file->d_name;
std::string cur_file_name(p_file->d_name);
file_names.push_back(cur_file_name);
}
}
closedir(p_dir);
return 0;
}
#endif

View File

@ -1,36 +1,24 @@
import torch
from torch import nn
import torchvision
import os
import sys
import struct
from torchsummary import summary
def main():
print('cuda device count: ', torch.cuda.device_count())
net = torch.load('ori_unet.pth')
net = net.to('cuda:0')
net = net.eval()
print('model: ', net)
#print('state dict: ', net.state_dict().keys())
tmp = torch.ones(1, 3, 224, 224).to('cuda:0')
print('input: ', tmp)
out = net(tmp)
device = torch.device('cpu')
state_dict = torch.load(sys.argv[1], map_location=device)
print('output:', out)
summary(net, (3, 224, 224))
#return
f = open("unet.wts", 'w')
f.write("{}\n".format(len(net.state_dict().keys())))
for k,v in net.state_dict().items():
print('key: ', k)
print('value: ', v.shape)
vr = v.reshape(-1).cpu().numpy()
f.write("{} {}".format(k, len(vr)))
for vv in vr:
f.write(" ")
f.write(struct.pack(">f", float(vv)).hex())
f.write("\n")
f = open("unet.wts", 'w')
f.write("{}\n".format(len(state_dict.keys())))
for k, v in state_dict.items():
print('key: ', k)
print('value: ', v.shape)
vr = v.reshape(-1).cpu().numpy()
f.write("{} {}".format(k, len(vr)))
for vv in vr:
f.write(" ")
f.write(struct.pack(">f", float(vv)).hex())
f.write("\n")
f.close()
if __name__ == '__main__':
main()
main()

View File

@ -4,14 +4,12 @@
#include "logging.h"
#include "common.hpp"
#define DEVICE 0
// #define USE_FP16 // comment out this if want to use FP16
#define USE_FP32 // USE_FP32 or USE_FP16
#define CONF_THRESH 0.5
#define BATCH_SIZE 1
#define cls 2
#define BILINEAR false
using namespace nvinfer1;
// stuff we know about the network and the input/output blobs
static const int INPUT_H = 640;
@ -21,362 +19,283 @@ const char* INPUT_BLOB_NAME = "data";
const char* OUTPUT_BLOB_NAME = "prob";
static Logger gLogger;
cv::Mat preprocess_img(cv::Mat& img) {
int w, h, x, y;
float r_w = INPUT_W / (img.cols * 1.0);
float r_h = INPUT_H / (img.rows * 1.0);
if (r_h > r_w) {
w = INPUT_W;
h = r_w * img.rows;
x = 0;
y = (INPUT_H - h) / 2;
}
else {
w = r_h * img.cols;
h = INPUT_H;
x = (INPUT_W - w) / 2;
y = 0;
}
cv::Mat re(h, w, CV_8UC3);
cv::resize(img, re, re.size(), 0, 0, cv::INTER_CUBIC);
cv::Mat out(INPUT_H, INPUT_W, CV_8UC3, cv::Scalar(128, 128, 128));
re.copyTo(out(cv::Rect(x, y, re.cols, re.rows)));
return out;
}
using namespace nvinfer1;
ILayer* doubleConv(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int outch, int ksize, std::string lname, int midch) {
Weights emptywts{ DataType::kFLOAT, nullptr, 0 };
// int p = ksize / 2;
// if (midch==NULL){
// midch = outch;
// }
IConvolutionLayer* conv1 = network->addConvolutionNd(input, midch, DimsHW{ ksize, ksize }, weightMap[lname + ".double_conv.0.weight"], emptywts);
conv1->setStrideNd(DimsHW{ 1, 1 });
conv1->setPaddingNd(DimsHW{ 1, 1 });
conv1->setNbGroups(1);
IScaleLayer* bn1 = addBatchNorm2d(network, weightMap, *conv1->getOutput(0), lname + ".double_conv.1", 0);
IActivationLayer* relu1 = network->addActivation(*bn1->getOutput(0), ActivationType::kLEAKY_RELU);
IConvolutionLayer* conv2 = network->addConvolutionNd(*relu1->getOutput(0), outch, DimsHW{ 3, 3 }, weightMap[lname + ".double_conv.3.weight"], emptywts);
conv2->setStrideNd(DimsHW{ 1, 1 });
conv2->setPaddingNd(DimsHW{ 1, 1 });
conv2->setNbGroups(1);
IScaleLayer* bn2 = addBatchNorm2d(network, weightMap, *conv2->getOutput(0), lname + ".double_conv.4", 0);
IActivationLayer* relu2 = network->addActivation(*bn2->getOutput(0), ActivationType::kLEAKY_RELU);
assert(relu2);
return relu2;
Weights emptywts{ DataType::kFLOAT, nullptr, 0 };
IConvolutionLayer* conv1 = network->addConvolutionNd(input, midch, DimsHW{ ksize, ksize }, weightMap[lname + ".double_conv.0.weight"], emptywts);
conv1->setStrideNd(DimsHW{ 1, 1 });
conv1->setPaddingNd(DimsHW{ 1, 1 });
conv1->setNbGroups(1);
IScaleLayer* bn1 = addBatchNorm2d(network, weightMap, *conv1->getOutput(0), lname + ".double_conv.1", 0);
IActivationLayer* relu1 = network->addActivation(*bn1->getOutput(0), ActivationType::kLEAKY_RELU);
IConvolutionLayer* conv2 = network->addConvolutionNd(*relu1->getOutput(0), outch, DimsHW{ 3, 3 }, weightMap[lname + ".double_conv.3.weight"], emptywts);
conv2->setStrideNd(DimsHW{ 1, 1 });
conv2->setPaddingNd(DimsHW{ 1, 1 });
conv2->setNbGroups(1);
IScaleLayer* bn2 = addBatchNorm2d(network, weightMap, *conv2->getOutput(0), lname + ".double_conv.4", 0);
IActivationLayer* relu2 = network->addActivation(*bn2->getOutput(0), ActivationType::kLEAKY_RELU);
assert(relu2);
return relu2;
}
ILayer* down(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int outch, int p, std::string lname) {
IPoolingLayer* pool1 = network->addPoolingNd(input, PoolingType::kMAX, DimsHW{ 2, 2 });
assert(pool1);
ILayer* dcov1 = doubleConv(network, weightMap, *pool1->getOutput(0), outch, 3, lname + ".maxpool_conv.1", outch);
assert(dcov1);
return dcov1;
IPoolingLayer* pool1 = network->addPoolingNd(input, PoolingType::kMAX, DimsHW{ 2, 2 });
assert(pool1);
ILayer* dcov1 = doubleConv(network, weightMap, *pool1->getOutput(0), outch, 3, lname + ".maxpool_conv.1", outch);
assert(dcov1);
return dcov1;
}
ILayer* up(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input1, ITensor& input2, int resize, int outch, int midch, std::string lname) {
float* deval = reinterpret_cast<float*>(malloc(sizeof(float) * resize * 2 * 2));
for (int i = 0; i < resize * 2 * 2; i++) {
deval[i] = 1.0;
}
if (BILINEAR) {
// add upsample bilinear
IResizeLayer* deconv1 = network->addResize(input1);
auto outdims = input2.getDimensions();
deconv1->setOutputDimensions(outdims);
deconv1->setResizeMode(ResizeMode::kLINEAR);
deconv1->setAlignCorners(true);
if (BILINEAR) {
// add upsample bilinear
IResizeLayer* deconv1 = network->addResize(input1);
auto outdims = input2.getDimensions();
deconv1->setOutputDimensions(outdims);
deconv1->setResizeMode(ResizeMode::kLINEAR);
deconv1->setAlignCorners(true);
int diffx = input2.getDimensions().d[1] - deconv1->getOutput(0)->getDimensions().d[1];
int diffy = input2.getDimensions().d[2] - deconv1->getOutput(0)->getDimensions().d[2];
int diffx = input2.getDimensions().d[1] - deconv1->getOutput(0)->getDimensions().d[1];
int diffy = input2.getDimensions().d[2] - deconv1->getOutput(0)->getDimensions().d[2];
ILayer* pad1 = network->addPaddingNd(*deconv1->getOutput(0), DimsHW{ diffx / 2, diffy / 2 }, DimsHW{ diffx - (diffx / 2), diffy - (diffy / 2) });
// dcov1->setPaddingNd(DimsHW{diffx / 2, diffx - diffx / 2},DimsHW{diffy / 2, diffy - diffy / 2});
ITensor* inputTensors[] = { &input2,pad1->getOutput(0) };
auto cat = network->addConcatenation(inputTensors, 2);
assert(cat);
if (midch == 64) {
ILayer* dcov1 = doubleConv(network, weightMap, *cat->getOutput(0), outch, 3, lname + ".conv", outch);
assert(dcov1);
return dcov1;
}
else {
int midch1 = outch / 2;
ILayer* dcov1 = doubleConv(network, weightMap, *cat->getOutput(0), midch1, 3, lname + ".conv", outch);
assert(dcov1);
return dcov1;
}
}
else {
/*Weights emptywts{ DataType::kFLOAT, nullptr, 0 };
Weights deconvwts1{ DataType::kFLOAT, deval, resize * 2 * 2 };*/
// weightMap[lname + ".up.weight"], weightMap[lname + ".up.bias"]
IDeconvolutionLayer* deconv1 = network->addDeconvolutionNd(input1, resize, DimsHW{ 2, 2 }, weightMap[lname + ".up.weight"], weightMap[lname + ".up.bias"]);
deconv1->setStrideNd(DimsHW{ 2, 2 });
deconv1->setNbGroups(1);
//weightMap["deconvwts." + lname] = deconvwts1;
int diffx = input2.getDimensions().d[1] - deconv1->getOutput(0)->getDimensions().d[1];
int diffy = input2.getDimensions().d[2] - deconv1->getOutput(0)->getDimensions().d[2];
ILayer* pad1 = network->addPaddingNd(*deconv1->getOutput(0), DimsHW{ diffx / 2, diffy / 2 }, DimsHW{ diffx - (diffx / 2), diffy - (diffy / 2) });
// dcov1->setPaddingNd(DimsHW{diffx / 2, diffx - diffx / 2},DimsHW{diffy / 2, diffy - diffy / 2});
ITensor* inputTensors[] = { &input2,pad1->getOutput(0) };
auto cat = network->addConcatenation(inputTensors, 2);
assert(cat);
ILayer* dcov1 = doubleConv(network, weightMap, *cat->getOutput(0), midch, 3, lname + ".conv", outch);
assert(dcov1);
return dcov1;
}
ILayer* pad1 = network->addPaddingNd(*deconv1->getOutput(0), DimsHW{ diffx / 2, diffy / 2 }, DimsHW{ diffx - (diffx / 2), diffy - (diffy / 2) });
// dcov1->setPaddingNd(DimsHW{diffx / 2, diffx - diffx / 2},DimsHW{diffy / 2, diffy - diffy / 2});
ITensor* inputTensors[] = { &input2,pad1->getOutput(0) };
auto cat = network->addConcatenation(inputTensors, 2);
assert(cat);
if (midch == 64) {
ILayer* dcov1 = doubleConv(network, weightMap, *cat->getOutput(0), outch, 3, lname + ".conv", outch);
assert(dcov1);
return dcov1;
} else {
int midch1 = outch / 2;
ILayer* dcov1 = doubleConv(network, weightMap, *cat->getOutput(0), midch1, 3, lname + ".conv", outch);
assert(dcov1);
return dcov1;
}
} else {
IDeconvolutionLayer* deconv1 = network->addDeconvolutionNd(input1, resize, DimsHW{ 2, 2 }, weightMap[lname + ".up.weight"], weightMap[lname + ".up.bias"]);
deconv1->setStrideNd(DimsHW{ 2, 2 });
deconv1->setNbGroups(1);
int diffx = input2.getDimensions().d[1] - deconv1->getOutput(0)->getDimensions().d[1];
int diffy = input2.getDimensions().d[2] - deconv1->getOutput(0)->getDimensions().d[2];
ILayer* pad1 = network->addPaddingNd(*deconv1->getOutput(0), DimsHW{ diffx / 2, diffy / 2 }, DimsHW{ diffx - (diffx / 2), diffy - (diffy / 2) });
// dcov1->setPaddingNd(DimsHW{diffx / 2, diffx - diffx / 2},DimsHW{diffy / 2, diffy - diffy / 2});
ITensor* inputTensors[] = { &input2,pad1->getOutput(0) };
auto cat = network->addConcatenation(inputTensors, 2);
assert(cat);
ILayer* dcov1 = doubleConv(network, weightMap, *cat->getOutput(0), midch, 3, lname + ".conv", outch);
assert(dcov1);
return dcov1;
}
}
ILayer* outConv(INetworkDefinition* network, std::map<std::string, Weights>& weightMap, ITensor& input, int outch, std::string lname) {
// Weights emptywts{DataType::kFLOAT, nullptr, 0};
IConvolutionLayer* conv1 = network->addConvolutionNd(input, cls, DimsHW{ 1, 1 }, weightMap[lname + ".conv.weight"], weightMap[lname + ".conv.bias"]);
assert(conv1);
conv1->setStrideNd(DimsHW{ 1, 1 });
conv1->setPaddingNd(DimsHW{ 0, 0 });
conv1->setNbGroups(1);
return conv1;
// Weights emptywts{DataType::kFLOAT, nullptr, 0};
IConvolutionLayer* conv1 = network->addConvolutionNd(input, cls, DimsHW{ 1, 1 }, weightMap[lname + ".conv.weight"], weightMap[lname + ".conv.bias"]);
assert(conv1);
conv1->setStrideNd(DimsHW{ 1, 1 });
conv1->setPaddingNd(DimsHW{ 0, 0 });
conv1->setNbGroups(1);
return conv1;
}
ICudaEngine* createEngine(unsigned int maxBatchSize, IBuilder* builder, IBuilderConfig* config, DataType dt, std::string wts_path) {
INetworkDefinition* network = builder->createNetworkV2(0U);
ICudaEngine* createEngine(unsigned int maxBatchSize, IBuilder* builder, IBuilderConfig* config, DataType dt, std::string wtsPath) {
INetworkDefinition* network = builder->createNetworkV2(0U);
// Create input tensor of shape {3, INPUT_H, INPUT_W} with name INPUT_BLOB_NAME
ITensor* data = network->addInput(INPUT_BLOB_NAME, dt, Dims3{ 3, INPUT_H, INPUT_W });
assert(data);
// Create input tensor of shape {3, INPUT_H, INPUT_W} with name INPUT_BLOB_NAME
ITensor* data = network->addInput(INPUT_BLOB_NAME, dt, Dims3{ 3, INPUT_H, INPUT_W });
assert(data);
std::map<std::string, Weights> weightMap = loadWeights(wts_path);
Weights emptywts{ DataType::kFLOAT, nullptr, 0 };
std::map<std::string, Weights> weightMap = loadWeights(wtsPath);
Weights emptywts{ DataType::kFLOAT, nullptr, 0 };
// build network
auto x1 = doubleConv(network, weightMap, *data, 64, 3, "inc", 64);
auto x2 = down(network, weightMap, *x1->getOutput(0), 128, 1, "down1");
auto x3 = down(network, weightMap, *x2->getOutput(0), 256, 1, "down2");
auto x4 = down(network, weightMap, *x3->getOutput(0), 512, 1, "down3");
auto channel = 512;
if (!BILINEAR) {
channel = 1024;
}
auto x5 = down(network, weightMap, *x4->getOutput(0), channel, 1, "down4");
ILayer* x6 = up(network, weightMap, *x5->getOutput(0), *x4->getOutput(0), 512, 512, 512, "up1");
ILayer* x7 = up(network, weightMap, *x6->getOutput(0), *x3->getOutput(0), 256, 256, 256, "up2");
ILayer* x8 = up(network, weightMap, *x7->getOutput(0), *x2->getOutput(0), 128, 128, 128, "up3");
ILayer* x9 = up(network, weightMap, *x8->getOutput(0), *x1->getOutput(0), 64, 64, 64, "up4");
ILayer* x10 = outConv(network, weightMap, *x9->getOutput(0), OUTPUT_SIZE, "outc");
// build network
auto x1 = doubleConv(network, weightMap, *data, 64, 3, "inc", 64);
auto x2 = down(network, weightMap, *x1->getOutput(0), 128, 1, "down1");
auto x3 = down(network, weightMap, *x2->getOutput(0), 256, 1, "down2");
auto x4 = down(network, weightMap, *x3->getOutput(0), 512, 1, "down3");
auto channel = 512;
if (!BILINEAR)
{
channel = 1024;
}
auto x5 = down(network, weightMap, *x4->getOutput(0), channel, 1, "down4");
ILayer* x6 = up(network, weightMap, *x5->getOutput(0), *x4->getOutput(0), 512, 512, 512, "up1");
ILayer* x7 = up(network, weightMap, *x6->getOutput(0), *x3->getOutput(0), 256, 256, 256, "up2");
ILayer* x8 = up(network, weightMap, *x7->getOutput(0), *x2->getOutput(0), 128, 128, 128, "up3");
ILayer* x9 = up(network, weightMap, *x8->getOutput(0), *x1->getOutput(0), 64, 64, 64, "up4");
ILayer* x10 = outConv(network, weightMap, *x9->getOutput(0), OUTPUT_SIZE, "outc");
x10->getOutput(0)->setName(OUTPUT_BLOB_NAME);
network->markOutput(*x10->getOutput(0));
std::cout << "set name out" << std::endl;
x10->getOutput(0)->setName(OUTPUT_BLOB_NAME);
network->markOutput(*x10->getOutput(0));
// Build engine
builder->setMaxBatchSize(maxBatchSize);
config->setMaxWorkspaceSize(16 * (1 << 20)); // 16MB
// Build engine
builder->setMaxBatchSize(maxBatchSize);
config->setMaxWorkspaceSize(16 * (1 << 20)); // 16MB
#ifdef USE_FP16
config->setFlag(BuilderFlag::kFP16);
config->setFlag(BuilderFlag::kFP16);
#endif
std::cout << "Building engine, please wait for a while..." << std::endl;
ICudaEngine* engine = builder->buildEngineWithConfig(*network, *config);
std::cout << "Build engine successfully!" << std::endl;
std::cout << "Building engine, please wait for a while..." << std::endl;
ICudaEngine* engine = builder->buildEngineWithConfig(*network, *config);
std::cout << "Build engine successfully!" << std::endl;
// Don't need the network any more
network->destroy();
// Don't need the network any more
network->destroy();
// Release host memory
for (auto& mem : weightMap)
{
free((void*)(mem.second.values));
}
// Release host memory
for (auto& mem : weightMap) {
free((void*)(mem.second.values));
}
return engine;
return engine;
}
void APIToModel(unsigned int maxBatchSize, IHostMemory** modelStream, std::string wtsPath) {
// Create builder
IBuilder* builder = createInferBuilder(gLogger);
IBuilderConfig* config = builder->createBuilderConfig();
void APIToModel(unsigned int maxBatchSize, IHostMemory** model_stream, std::string wts_path) {
// Create builder
IBuilder* builder = createInferBuilder(gLogger);
IBuilderConfig* config = builder->createBuilderConfig();
// Create model to populate the network, then set the outputs and create an engine
// ICudaEngine* engine = (CREATENET(NET))(maxBatchSize, builder, config, DataType::kFLOAT);
ICudaEngine* engine = createEngine(maxBatchSize, builder, config, DataType::kFLOAT, wtsPath);
assert(engine != nullptr);
// Create model to populate the network, then set the outputs and create an engine
ICudaEngine* engine = createEngine(maxBatchSize, builder, config, DataType::kFLOAT, wts_path);
assert(engine != nullptr);
// Serialize the engine
(*modelStream) = engine->serialize();
// Serialize the engine
(*model_stream) = engine->serialize();
// Close everything down
engine->destroy();
builder->destroy();
// Close everything down
engine->destroy();
builder->destroy();
}
void doInference(IExecutionContext& context, float* input, float* output, int batchSize) {
const ICudaEngine& engine = context.getEngine();
const ICudaEngine& engine = context.getEngine();
// Pointers to input and output device buffers to pass to engine.
// Engine requires exactly IEngine::getNbBindings() number of buffers.
assert(engine.getNbBindings() == 2);
void* buffers[2];
// Pointers to input and output device buffers to pass to engine.
// Engine requires exactly IEngine::getNbBindings() number of buffers.
assert(engine.getNbBindings() == 2);
void* buffers[2];
// In order to bind the buffers, we need to know the names of the input and output tensors.
// Note that indices are guaranteed to be less than IEngine::getNbBindings()
const int inputIndex = engine.getBindingIndex(INPUT_BLOB_NAME);
const int outputIndex = engine.getBindingIndex(OUTPUT_BLOB_NAME);
// In order to bind the buffers, we need to know the names of the input and output tensors.
// Note that indices are guaranteed to be less than IEngine::getNbBindings()
const int inputIndex = engine.getBindingIndex(INPUT_BLOB_NAME);
const int outputIndex = engine.getBindingIndex(OUTPUT_BLOB_NAME);
// Create GPU buffers on device
CHECK(cudaMalloc(&buffers[inputIndex], batchSize * 3 * INPUT_H * INPUT_W * sizeof(float)));
CHECK(cudaMalloc(&buffers[outputIndex], batchSize * OUTPUT_SIZE * sizeof(float)));
// Create GPU buffers on device
CHECK(cudaMalloc(&buffers[inputIndex], batchSize * 3 * INPUT_H * INPUT_W * sizeof(float)));
CHECK(cudaMalloc(&buffers[outputIndex], batchSize * OUTPUT_SIZE * sizeof(float)));
// Create stream
cudaStream_t stream;
CHECK(cudaStreamCreate(&stream));
// Create stream
cudaStream_t stream;
CHECK(cudaStreamCreate(&stream));
// DMA input batch data to device, infer on the batch asynchronously, and DMA output back to host
CHECK(cudaMemcpyAsync(buffers[inputIndex], input, batchSize * 3 * INPUT_H * INPUT_W * sizeof(float), cudaMemcpyHostToDevice, stream));
context.enqueue(batchSize, buffers, stream, nullptr);
CHECK(cudaMemcpyAsync(output, buffers[outputIndex], batchSize * OUTPUT_SIZE * sizeof(float), cudaMemcpyDeviceToHost, stream));
// DMA input batch data to device, infer on the batch asynchronously, and DMA output back to host
CHECK(cudaMemcpyAsync(buffers[inputIndex], input, batchSize * 3 * INPUT_H * INPUT_W * sizeof(float), cudaMemcpyHostToDevice, stream));
context.enqueue(batchSize, buffers, stream, nullptr);
CHECK(cudaMemcpyAsync(output, buffers[outputIndex], batchSize * OUTPUT_SIZE * sizeof(float), cudaMemcpyDeviceToHost, stream));
cudaStreamSynchronize(stream);
cudaStreamSynchronize(stream);
// Release stream and buffers
cudaStreamDestroy(stream);
CHECK(cudaFree(buffers[inputIndex]));
CHECK(cudaFree(buffers[outputIndex]));
}
struct Detection {
float mask[INPUT_W * INPUT_H * 1];
};
float sigmoid(float x) {
return (1 / (1 + exp(-x)));
}
void process_cls_result(Detection& res, float* output) {
for (int i = 0; i < INPUT_W * INPUT_H * 1; i++) {
res.mask[i] = sigmoid(*(output + i));
}
// Release stream and buffers
cudaStreamDestroy(stream);
CHECK(cudaFree(buffers[inputIndex]));
CHECK(cudaFree(buffers[outputIndex]));
}
int main(int argc, char** argv) {
cudaSetDevice(DEVICE);
// create a model using the API directly and serialize it to a stream
char* trtModelStream{ nullptr };
size_t size{ 0 };
std::string engine_name = "unet.engine";
std::vector<std::string> file_names;
std::string wtsPath = "..\\models\\unet_carvana_scale0.5_epoch2.wts";
if (argc == 2 && std::string(argv[1]) == "-s") {
IHostMemory* modelStream{ nullptr };
APIToModel(BATCH_SIZE, &modelStream, wtsPath);
assert(modelStream != nullptr);
std::ofstream p(engine_name, std::ios::binary);
if (!p) {
std::cerr << "could not open plan output file" << std::endl;
return -1;
}
p.write(reinterpret_cast<const char*>(modelStream->data()), modelStream->size());
modelStream->destroy();
return 0;
}
else if (argc == 3 && std::string(argv[1]) == "-d") {
std::ifstream file(engine_name, std::ios::binary);
if (file.good()) {
file.seekg(0, file.end);
size = file.tellg();
file.seekg(0, file.beg);
trtModelStream = new char[size];
assert(trtModelStream);
file.read(trtModelStream, size);
file.close();
cv::glob(argv[2], file_names);
}
}
else {
std::cerr << "arguments not right!" << std::endl;
std::cerr << "./unet -s // serialize model to plan file" << std::endl;
std::cerr << "./unet -d ../samples // deserialize plan file and run inference" << std::endl;
return -1;
}
cudaSetDevice(DEVICE);
//std::vector<std::string> file_names;
//if (read_files_in_dir(argv[2], file_names) < 0) {
// std::cout << "read_files_in_dir failed." << std::endl;
// return -1;
//}
char* trt_model_stream = nullptr;
size_t size = 0;
std::string engine_name = "unet.engine";
std::string wts_path = "unet.wts";
// prepare input data ---------------------------
static float data[BATCH_SIZE * 3 * INPUT_H * INPUT_W];
//for (int i = 0; i < 3 * INPUT_H * INPUT_W; i++)
// data[i] = 1.0;
static float prob[BATCH_SIZE * OUTPUT_SIZE];
IRuntime* runtime = createInferRuntime(gLogger);
assert(runtime != nullptr);
ICudaEngine* engine = runtime->deserializeCudaEngine(trtModelStream, size);
assert(engine != nullptr);
IExecutionContext* context = engine->createExecutionContext();
assert(context != nullptr);
delete[] trtModelStream;
if (argc == 2 && std::string(argv[1]) == "-s") {
// Create a TensorRT model and serialize it to a file
IHostMemory* model_stream{ nullptr };
APIToModel(BATCH_SIZE, &model_stream, wts_path);
assert(model_stream != nullptr);
std::ofstream p(engine_name, std::ios::binary);
if (!p) {
std::cerr << "could not open plan output file" << std::endl;
return -1;
}
p.write(reinterpret_cast<const char*>(model_stream->data()), model_stream->size());
model_stream->destroy();
return 0;
} else if (argc == 3 && std::string(argv[1]) == "-d") {
// Load engine file
std::ifstream file(engine_name, std::ios::binary);
if (file.good()) {
file.seekg(0, file.end);
size = file.tellg();
file.seekg(0, file.beg);
trt_model_stream = new char[size];
assert(trt_model_stream);
file.read(trt_model_stream, size);
file.close();
}
} else {
std::cerr << "arguments not right!" << std::endl;
std::cerr << "./unet -s // serialize model to plan file" << std::endl;
std::cerr << "./unet -d ../samples // deserialize plan file and run inference" << std::endl;
return -1;
}
// Prepare input output data
static float data[BATCH_SIZE * 3 * INPUT_H * INPUT_W];
static float prob[BATCH_SIZE * OUTPUT_SIZE];
cv::Mat results = cv::Mat::zeros(INPUT_H, INPUT_W, CV_8UC3);
for (int f = 0; f < (int)file_names.size(); f++)
{
// Deserialize engine
IRuntime* runtime = createInferRuntime(gLogger);
assert(runtime != nullptr);
ICudaEngine* engine = runtime->deserializeCudaEngine(trt_model_stream, size);
assert(engine != nullptr);
IExecutionContext* context = engine->createExecutionContext();
assert(context != nullptr);
delete[] trt_model_stream;
cv::Mat img = cv::imread(file_names[f]);
if (img.empty()) continue;
cv::Mat pr_img = preprocess_img(img);
//cv::resize(img, pr_img, cv::Size(INPUT_W, INPUT_H));
cv::Mat img = cv::imread(argv[2]);
for (int i = 0; i < INPUT_H * INPUT_W; i++) {
data[i] = (pr_img.at<cv::Vec3b>(i)[2]) / 255.0;
data[i + INPUT_H * INPUT_W] = (pr_img.at<cv::Vec3b>(i)[1]) / 255.0;
data[i + 2 * INPUT_H * INPUT_W] = (pr_img.at<cv::Vec3b>(i)[0]) / 255.0;
}
// Run inference
auto start = std::chrono::system_clock::now();
doInference(*context, data, prob, BATCH_SIZE);
auto end = std::chrono::system_clock::now();
std::cout << std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count() << "ms" << std::endl;
for (int i = 0; i < INPUT_H * INPUT_W; i++) {
float fmax = 0.0;
int index = 0;
for (int j = 0; j < cls; j++) {
if (prob[i + j * INPUT_H * INPUT_W] > fmax) {
index = j;
fmax = prob[i + j * INPUT_H * INPUT_W];
}
}
// Preprocess
cv::resize(img, img, cv::Size(INPUT_W, INPUT_H));
for (int i = 0; i < INPUT_H * INPUT_W; i++) {
data[i] = (img.at<cv::Vec3b>(i)[2]) / 255.0;
data[i + INPUT_H * INPUT_W] = (img.at<cv::Vec3b>(i)[1]) / 255.0;
data[i + 2 * INPUT_H * INPUT_W] = (img.at<cv::Vec3b>(i)[0]) / 255.0;
}
if (index == 1) {
results.at<cv::Vec3b>(i) = cv::Vec3b(255, 255, 255);
}
// Run inference
auto start = std::chrono::system_clock::now();
doInference(*context, data, prob, BATCH_SIZE);
auto end = std::chrono::system_clock::now();
std::cout << std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count() << "ms" << std::endl;
else {
results.at<cv::Vec3b>(i) = cv::Vec3b(0, 0, 0);
}
}
cv::imshow(" results", results);
cv::imwrite(f + "_unet.jpg", results);
// Postprocess
cv::Mat result = cv::Mat::zeros(INPUT_H, INPUT_W, CV_8UC3);
for (int i = 0; i < INPUT_H * INPUT_W; i++) {
float fmax = prob[i];
int index = 0;
for (int j = 1; j < cls; j++) {
if (prob[i + j * INPUT_H * INPUT_W] > fmax) {
index = j;
fmax = prob[i + j * INPUT_H * INPUT_W];
}
}
cv::waitKey(0);
results = cv::Mat::zeros(INPUT_H, INPUT_W, CV_8UC3);
if (index == 1) {
result.at<cv::Vec3b>(i) = cv::Vec3b(255, 255, 255);
}
}
}
cv::imwrite("result.jpg", result);
// Destroy the engine
context->destroy();
engine->destroy();
runtime->destroy();
// Destroy the engine
context->destroy();
engine->destroy();
runtime->destroy();
return 0;
return 0;
}